"""
Internal base utilities.
"""

from typing import Optional, Tuple, Any, Sequence
from json.decoder import JSONDecodeError

import time
import requests
from .exceptions import BaseXToolsException, NotFound, TooManyEdits

BASE_URL = "https://xtools.wmflabs.org/api"
START_TIME = "2000-01-01"
END_TIME = "2050-12-31"


def url(path: str) -> str:
    """
    Return a full API URL.
    :param path:
    :return:
    """
    return BASE_URL + path


def error_exception(response: dict) -> Optional[BaseXToolsException]:
    """
    Return an optional Exception instance for an error response. Return ``None`` if the given response is not an error.
    :param response: response from the API.
    :return: ``None`` or a ``BaseXToolsException`` (or subclass) instance.
    """
    if "error" not in response:
        return

    error = response["error"]

    # 'error' can be a dict or a string
    if isinstance(error, dict):
        code = error["code"]
        if code == 404:
            return NotFound(error["message"])

        return BaseXToolsException(error["message"], code)

    if error.startswith("No matching result") \
            or error.endswith("is not a valid project") \
            or error == "The requested user does not exist":
        return NotFound(error)

    if error.startswith("User has made too many edits!"):
        return TooManyEdits(error)

    return BaseXToolsException(str(error))


def get(path: str, params=None, retry=3, retry_delay=1):
    """
    Perform a GET request against the API.
    :param path:
    :param params:
    :param retry:
    :param retry_delay:
    :return:
    """
    r = requests.get(url(path), params=params)
    # 'Proxy error'
    if r.status_code == 502:
        if retry > 0:
            time.sleep(retry_delay)
            return get(path, params, retry - 1)
        r.raise_for_status()

    try:
        response = r.json()
    except JSONDecodeError as e:
        r.raise_for_status()
        raise e

    exception = error_exception(response)
    if exception:
        raise exception
    return response


def build_path(path_format: str, params: Sequence[Tuple[str, Any, str]]) -> str:
    """
    Build a path for the XTools API.

    Examples:

        >>> build_path("/{a}", [("a", "foo", "bar")])
        "/foo"
        >>> build_path("/{a}", [("a", "", "bar")])
        "/bar"
        >>> build_path("/{a}/{b}/{c}", [("a", "", "x"), ("b", "B", "x"), ("c", "", "x")])
        "/x/B"

    :param path_format:
    :param params:
    :return:
    """
    params_dict = {}

    def has_more(index):
        for _, val, _ in params[index+1:]:
            if val:
                return True
        return False

    for i, (name, value, default) in enumerate(params):
        if value:
            param_value = str(value)
        else:
            param_value = ""

            if has_more(i):
                if default:
                    param_value = default
                # safe defaults
                elif name == "start":
                    param_value = START_TIME
                elif name == "end":
                    param_value = END_TIME

        params_dict[name] = param_value

    path = path_format.format(**params_dict).rstrip("/")
    return path


"""
Endpoints related to pages.

https://xtools.readthedocs.io/en/stable/api/page.html
"""

from typing import Optional, Sequence
from datetime import date

from . import base


def _get_page_dict(what: str, project: str, article: str) -> dict:
    return base.get("/page/%s/%s/%s" % (what, project, article))


def article_info(project: str, article: str) -> dict:
    """
    Return basic information about an article, such as page views, watchers, edits counts; author; assessment.

    https://xtools.readthedocs.io/en/stable/api/page.html#article-info

    :param project:
    :param article:
    :return:
    """
    return _get_page_dict("articleinfo", project, article)


def prose(project: str, article: str) -> dict:
    """
    Return prose information about an article, such as references and words counts.

    https://xtools.readthedocs.io/en/stable/api/page.html#prose

    Example:

        >>> prose("en.wikipedia.org", "Albert_Einstein")
        {
            "project": "en.wikipedia.org",
            "page": "Albert_Einstein",
            "characters": 66690,
            "words": 10536,
            "references": 308,
            "unique_references": 251,
            "sections": 63,
            "elapsed_time": 0.645,
        }

    :param project:
    :param article:
    :return:
    """
    return _get_page_dict("prose", project, article)


def links(project: str, article: str) -> dict:
    """
    Return in and outgoing links counts of an article.

    Example:

        >>> links("enwiki", "Albert_Einstein")
        {
          "project": "en.wikipedia.org",
          "page": "Albert_Einstein",
          "links_ext_count": 372,
          "links_out_count": 1178,
          "links_in_count": 9744,
          "redirects_count": 19,
          "elapsed_time": 17.902
        }

    https://xtools.readthedocs.io/en/stable/api/page.html#links

    :param project:
    :param article:
    :return:
    """
    return _get_page_dict("links", project, article)


def top_editors(project: str, article: str,
                start: Optional[date] = None,
                end: Optional[date] = None,
                limit: Optional[int] = None,
                exclude_bots: bool = False) -> dict:
    """
    Get top editors on an article.

    Note: as of 2020/05/10, passing a limit without start and end gives an error.

    Example:

        >>> top_editors("en.wikipedia", "Albert_Einstein")
        {
          "project": "en.wikipedia.org",
          "page": "Albert_Einstein",
          "limit": 20,
          "top_editors": [
            {
              "rank": 1,
              "username": "Otterpops",
              "count": 509,
              "minor": 46,
              "first_edit": {"id": 111357663, "timestamp": 20070227165531},
              "latest_edit": {"id": 115188249, "timestamp": 20070314232325},
            },
            ...
          ],
          "elapsed_time": 0.133,
        }

    https://xtools.readthedocs.io/en/stable/api/page.html#top-editors

    :param project:
    :param article:
    :param start: optional start date.
    :param end: optional end date.
    :param limit: optional limit (XTools' default is 20)
    :param exclude_bots: if True, exclude bots.
    :return:
    """
    path = base.build_path("/page/top_editors/{project}/{article}/{start}/{end}/{limit}", (
        ("project", project, ""),
        ("article", article, ""),
        ("start", start, ""),
        ("end", end, ""),
        ("limit", limit, ""),
    ))
    params = {}
    if exclude_bots:
        params["nobots"] = "1"

    return base.get(path, params)


def assessments(project: str, articles: Sequence[str],
                class_only: bool = False) -> dict:
    """
    Get assessment data for the given articles.

    Example:

        >>> assessments("en.wikipedia.org", ["Albert_Einstein", "Bob_Dylan"], class_only=True)
        {
          "project": "en.wikipedia.org",
          "Albert Einstein": {
            "value": "GA",
            "color": "#66FF66",
            "category": "Category:GA-Class articles",
            "badge": "https://upload.wikimedia.org/wikipedia/commons/9/94/Symbol_support_vote.svg"
          },
          "Bob Dylan": {
            "value": "FA",
            "color": "#9CBDFF",
            "category": "Category:FA-Class articles",
            "badge": "https://upload.wikimedia.org/wikipedia/commons/b/bc/Featured_article_star.svg"
          },
          "elapsed_time": 0.361
        }

    https://xtools.readthedocs.io/en/stable/api/page.html#assessments

    :param project:
    :param articles:
    :param class_only:
    :return:
    """
    params = {}
    if class_only:
        params["classonly"] = "1"

    return base.get("/page/assessments/%s/%s" % (project, "|".join(articles)), params)


"""
Endpoints related to quotes. They all return either one Quote object or a sequence of it.

A Quote is a named tuple with ``id`` and ``text`` fields.

https://xtools.readthedocs.io/en/stable/api/quote.html
"""

from typing import Sequence
from collections import namedtuple

from . import base

Quote = namedtuple("Quote", ("id", "text"), module=__name__)


def _get_quotes(path: str) -> Sequence[Quote]:
    raw_quotes = base.get(path)
    quotes = []

    for quote_id, quote_text in sorted(raw_quotes.items()):
        quotes.append(Quote(int(quote_id), quote_text))

    return quotes


def random_quote() -> Quote:
    """
    Return a random quote.

    https://xtools.readthedocs.io/en/stable/api/quote.html#random-quote

    :return: Quote.
    """
    return _get_quotes("/quote/random")[0]


def single_quote(quote_id: int) -> Quote:
    """
    Get a quote by ID.

    https://xtools.readthedocs.io/en/stable/api/quote.html#single-quote

    :param quote_id:
    :return: Quote.
    """
    return _get_quotes("/quote/%d" % quote_id)[0]


def all_quotes() -> Sequence[Quote]:
    """
    Get all quotes.

    https://xtools.readthedocs.io/en/stable/api/quote.html#all-quotes

    :return: sequence of Quote objects.
    """
    return _get_quotes("/quote/all")


"""
Endpoints related to projects.

https://xtools.readthedocs.io/en/stable/api/project.html
"""

from typing import Optional, Sequence
from datetime import date

from . import base


def _get_project_dict(what: str, project: str) -> dict:
    return base.get("/project/%s/%s" % (what, project))


def normalize_project(project: str) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/project.html#normalize-project

    Example:

        >>> normalize_project("enwiki")
        {
          "project": "en.wikipedia.org",
          "domain": "en.wikipedia.org",
          "url": "https://en.wikipedia.org/",
          "api": "https://en.wikipedia.org/w/api.php",
          "database": "enwiki",
          "elapsed_time": 0.051
        }

    :param project:
    :return:
    """
    return _get_project_dict("normalize", project)


def namespaces(project: str) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/project.html#namespaces

    :param project:
    :return:
    """
    return _get_project_dict("namespaces", project)


def page_assessments(project: str) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/project.html#page-assessments

    :param project:
    :return:
    """
    return _get_project_dict("assessments", project)


def page_assessments_configuration() -> dict:
    """
    Return the list of wikis that support page assessments and the configuration for each.

    https://xtools.readthedocs.io/en/stable/api/project.html#page-assessments-configuration

    Example:

        >>> page_assessments_configuration()
        {
          "projects": [
            "ar.wikipedia.org",
            "en.wikipedia.org",
            "en.wikivoyage.org",
            ...
          ],
          "config": {
            "ar.wikipedia.org": { ... },
            ...
          },
        }

    :return:
    """
    return base.get("/project/assessments")


def automated_tools(project: str) -> dict:
    """
    Return a list of known (semi-)automated tools used on the project.

    Example:

        >>> automated_tools("en.wikipedia.org")
        {
          "project": "en.wikipedia.org",
          "Admin actions": {
            "regex": "^...",
            "link": "Project:Administrators",
            "label": "Admin actions",
            "namespaces": [],
            "tags": [],
          },
          "Advisor.js": {
            "regex": "...Advisor.js",
            ...
          },
          ...
        }

    Note the exact same endpoint exists under ``api/user``.

    https://xtools.readthedocs.io/en/stable/api/project.html#automated-tools
    https://xtools.readthedocs.io/en/stable/api/user.html#automated-tools

    :param project:
    :return:
    """
    return _get_project_dict("automated_tools", project)


def admins_and_user_groups(project: str) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/project.html#admins-and-user-groups

    :param project:
    :return:
    """
    return _get_project_dict("admins_groups", project)


def _get_project_stats_dict(what: str, project: str,
                            start: Optional[date] = None,
                            end: Optional[date] = None,
                            actions: Optional[Sequence[str]] = None) -> dict:
    path = base.build_path("/project/{what}/{project}/{start}/{end}", (
        ("what", what, ""),
        ("project", project, ""),
        ("start", start, ""),
        ("end", end, ""),
    ))

    params = {}
    if actions:
        params["actions"] = "|".join(actions)

    return base.get(path, params)


def admin_statistics(project: str,
                     start: Optional[date] = None,
                     end: Optional[date] = None,
                     actions: Optional[Sequence[str]] = None) -> dict:
    """
    Return admin users of the project with counts of the actions they took.

    XTools limits the time period to one month.

    https://xtools.readthedocs.io/en/stable/api/project.html#admin-statistics

    :param project:
    :param start:
    :param end:
    :param actions: available actions include: ``delete``, ``revision-delete``, ``log-delete``, ``restore``,
      ``re-block``, ``unblock``, ``re-protect``, ``unprotect``, ``rights``, ``merge``, ``import``.
    :return:
    """
    return _get_project_stats_dict("admin_stats", project, start, end, actions)


def patroller_statistics(project: str,
                         start: Optional[date] = None,
                         end: Optional[date] = None,
                         actions: Optional[Sequence[str]] = None) -> dict:
    """
    Same as ``admin_statistics`` with different actions.

    https://xtools.readthedocs.io/en/stable/api/project.html#patroller-statistics

    :param project: 
    :param start: 
    :param end: 
    :param actions: available actions include: ``patrol``, ``page-curation``, ``pc-accept``, ``pc-reject``.
    :return: 
    """
    return _get_project_stats_dict("patroller_stats", project, start, end, actions)


def steward_statistics(project: str,
                       start: Optional[date] = None,
                       end: Optional[date] = None,
                       actions: Optional[Sequence[str]] = None) -> dict:
    """
    Same as ``admin_statistics`` with different actions.

    https://xtools.readthedocs.io/en/stable/api/project.html#stewards-statistics

    :param project:
    :param start:
    :param end:
    :param actions: available actions include: ``global-account-un-lock`` (global locks and unlocks), ``global-block``,
      ``global-unblock``, ``global-rename``, ``global-rights``, ``wiki-set-change``.
    :return:
    """
    return _get_project_stats_dict("steward_stats", project, start, end, actions)


from typing import Optional

__all__ = ['BaseXToolsException', 'NotFound', 'TooManyEdits']


class BaseXToolsException(Exception):
    def __init__(self, message, code: Optional[int] = None):
        super().__init__(message)
        self.message = message
        self.code = code

    def __eq__(self, other):
        return isinstance(other, BaseXToolsException) \
            and str(self) == str(other) \
            and self.code == other.code


class NotFound(BaseXToolsException):
    pass


class TooManyEdits(BaseXToolsException):
    pass


"""
Endpoints related to users.

https://xtools.readthedocs.io/en/stable/api/user.html
"""

from typing import Optional, Generator, Sequence
from datetime import date

from . import base


def simple_edit_count(project: str, username: str,
                      namespace: Optional[int] = None,
                      start: Optional[date] = None,
                      end: Optional[date] = None) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/user.html#simple-edit-count

    Example:

        >>> simple_edit_count("en.wikipedia.org", "Jimbo_Wales")
        {
          "project": "en.wikipedia.org",
          "username": "Jimbo_Wales",
          "namespace": "all",
          "user_id": 24,
          "deleted_edit_count": 424,
          "live_edit_count": 13751,
          ...,
          "elapsed_time": 80.916
        }

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param start:
    :param end:
    :return:
    """

    path = base.build_path("/user/simple_editcount/{project}/{username}/{namespace}/{start}/{end}", (
        ("project", project, ""),
        ("username", username, ""),
        ("namespace", namespace, "0"),
        ("start", start, ""),
        ("end", end, ""),
    ))

    return base.get(path)


def number_of_pages_created(project: str, username: str,
                            namespace: Optional[str] = None,
                            redirects: Optional[str] = None,
                            deleted: Optional[str] = None,
                            start: Optional[date] = None,
                            end: Optional[date] = None) -> dict:
    """
    Return the number of pages created by a user along with some other info.

    Example:

        >>> number_of_pages_created("en.wikipedia", "Jimbo_Wales")
        {
          "project": "en.wikipedia.org",
          "username": "Jimbo_Wales",
          ...
          "counts": {
            "count": 42,
            "total_length": 30691,
            "avg_length": 730.7380952380952,
            "deleted": 1
          },
          "elapsed_time": 0.059,
        }

    https://xtools.readthedocs.io/en/stable/api/user.html#number-of-pages-created

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param redirects: either ``"noredirects"`` (default), ``"onlyredirects"``, or ``"all"`` for both.
    :param deleted: either ``"live"``, ``"deleted"``, or ``"all"`` (default).
    :param start:
    :param end:
    :return:
    """
    path = base.build_path("/user/pages_count/{project}/{username}/{namespace}/{redirects}/{deleted}/{start}/{end}", (
        ("project", project, ""),
        ("username", username, ""),
        ("namespace", namespace, "0"),
        ("redirects", redirects, "noredirects"),
        ("deleted", deleted, "all"),
        ("start", start, ""),
        ("end", end, ""),
    ))

    return base.get(path)


def _fix_page(page: dict) -> dict:
    """
    Fix a page dict as returned by the API.
    :param page:
    :return:
    """
    # Some (all?) pages with a number as their title have an int in this field.
    # E.g. https://fr.wikipedia.org/wiki/2040.
    if "page_title" in page:
        page["page_title"] = str(page["page_title"])
    return page


def pages_created(project: str, username: str,
                  namespace: str = None,
                  redirects: Optional[str] = None,
                  deleted: Optional[str] = None,
                  start: Optional[date] = None,
                  end: Optional[date] = None,
                  offset: Optional[int] = None,
                  all_times: bool = False) -> dict:
    """
    Return pages created by a user. This does not handle pagination; see ``pages_created_iter`` for that.

    Note that XTools doesn't support users with >400,000 edits, and this function raises
    a ``xtools.exceptions.TooManyEdits`` exception if you try.

    Example:

        >>> pages_created("enwiki", "Jimbo_Wales")
        {
          "pages": [
            {
              "namespace": 0,
              "type": "rev",
              "page_title": "Helen Crlenkovich",
              "page_is_redirect": 0,
              "length": 10691,
              "page_len": 11597,
              "rev_timestamp": 20200509210713,
              "rev_len": 10691,
              "rev_id": 170658958,
              "recreated": null,
              "pa_class": null,
              "pa_importance": null,
              "pa_page_revision": null,
              "raw_time": 20200509210713,
              "human_time": "2020-05-09 21:07",
              "badge": "https://upload.wikimedia.org/....svg",
            },
            ...
          ],
          "continue": 1,
        }

    https://xtools.readthedocs.io/en/stable/api/user.html#pages-created

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param redirects: either ``"noredirects"`` (default), ``"onlyredirects"``, or ``"all"`` for both.
    :param deleted: either ``"live"``, ``"deleted"``, or ``"all"`` (default).
    :param start:
    :param end:
    :param offset:
    :param all_times: if ``True``, fetch all pages instead of the more recent ones. This overrides ``start``
      and ``end``.
    :return:
    """
    if all_times:
        start = base.START_TIME
        end = base.END_TIME

    path = base.build_path(
        "/user/pages/{project}/{username}/{namespace}/{redirects}/{deleted}/{start}/{end}/{offset}",
        (
            ("project", project, ""),
            ("username", username, ""),
            ("namespace", namespace, "0"),
            ("redirects", redirects, "noredirects"),
            ("deleted", deleted, "all"),
            ("start", start, ""),
            ("end", end, ""),
            ("offset", offset, ""),
        ))

    ret = base.get(path)

    # Flatten the 'pages' key
    pages = ret["pages"]
    # single namespace: {"0": A, "1": B, ...}
    # all namespaces: {"0": [A, B], "1": [C, D, E], ...}
    keys = sorted(pages.keys(), key=int)

    is_namespaced = False
    try:
        is_namespaced = isinstance(next(pages.values().__iter__()), list)
    except StopIteration:
        pass

    if is_namespaced:
        ret["pages"] = [_fix_page(page) for k in keys for page in pages[k]]
    else:
        ret["pages"] = [_fix_page(pages[k]) for k in keys]

    return ret


def pages_created_iter(project: str, username: str,
                       namespace: str = None,
                       redirects: Optional[str] = None,
                       deleted: Optional[str] = None,
                       start: Optional[date] = None,
                       end: Optional[date] = None,
                       all_times: bool = False) -> Generator[dict, None, None]:
    """
    Equivalent of ``pages_created`` that yields page dicts and does the pagination for you.

    See ``pages_created`` for sample return values.

    Example:

        >>> for page in pages_created_iter("enwiki", "Jimbo_Wales"):
        ...     print(page["page_title"])

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param redirects: either ``"noredirects"`` (default), ``"onlyredirects"``, or ``"all"`` for both.
    :param deleted: either ``"live"``, ``"deleted"``, or ``"all"`` (default).
    :param start:
    :param end:
    :param all_times: if ``True``, fetch all pages instead of the more recent ones. This overrides ``start``
      and ``end``.
    :return:
    """
    offset = 0

    while offset is not None:
        ret = pages_created(project, username,
                            namespace=namespace, redirects=redirects, deleted=deleted,
                            start=start, end=end, all_times=all_times, offset=offset)
        offset = ret.get("continue")
        for page in ret["pages"]:
            yield page


def automated_edit_counter(project: str, username: str,
                           namespace: Optional[str] = None,
                           start: Optional[date] = None,
                           end: Optional[date] = None,
                           offset: Optional[int] = None,
                           tools: bool = False) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/user.html#automated-edit-counter

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param start:
    :param end:
    :param offset:
    :param tools:
    :return:
    """
    path = base.build_path(
        "/user/automated_editcount/{project}/{username}/{namespace}/{start}/{end}/{offset}/{tools}",
        (
            ("project", project, ""),
            ("username", username, ""),
            ("namespace", namespace, "0"),
            ("start", start, ""),
            ("end", end, ""),
            ("offset", offset, ""),
            ("tools", "true" if tools else "", ""),
        ))
    return base.get(path)


def _edits(what: str, project: str, username: str, namespace: str = "0",
           start: Optional[date] = None,
           end: Optional[date] = None,
           offset: Optional[int] = None) -> dict:
    """
    Base function for both ``non_automated_edits`` and ``automated_edits``.

    :param what:
    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param start:
    :param end:
    :param offset:
    :return:
    """
    path = base.build_path("/user/{what}/{project}/{username}/{namespace}/{start}/{end}/{offset}", (
        ("what", what, ""),
        ("project", project, ""),
        ("username", username, ""),
        ("namespace", namespace, "0"),
        ("start", start, ""),
        ("end", end, ""),
        ("offset", offset, ""),
    ))
    return base.get(path)


def non_automated_edits(project: str, username: str,
                        namespace: Optional[str] = None,
                        start: Optional[date] = None,
                        end: Optional[date] = None,
                        offset: Optional[int] = None) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/user.html#non-automated-edits

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param start:
    :param end:
    :param offset:
    :return:
    """
    return _edits("nonautomated_edits", project, username,
                  namespace=namespace, start=start, end=end, offset=offset)


def automated_edits(project: str, username: str,
                    namespace: Optional[str] = None,
                    start: Optional[date] = None,
                    end: Optional[date] = None,
                    offset: Optional[int] = None) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/user.html#automated-edits

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param start:
    :param end:
    :param offset:
    :return:
    """
    return _edits("automated_edits", project, username,
                  namespace=namespace, start=start, end=end, offset=offset)


def edit_summaries(project: str, username: str,
                   namespace: Optional[str] = None,
                   start: Optional[date] = None,
                   end: Optional[date] = None) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/user.html#edit-summaries

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param start:
    :param end:
    :return:
    """
    path = base.build_path("/user/edit_summaries/{project}/{username}/{namespace}/{start}/{end}", (
        ("project", project, ""),
        ("username", username, ""),
        ("namespace", namespace, "0"),
        ("start", start, ""),
        ("end", end, ""),
    ))
    return base.get(path)


def top_edits(project: str, username: str,
              namespace: Optional[str] = None,
              page_title: Optional[str] = None) -> dict:
    """
    Return the top-edited pages by a user, or all edits made by a user to a specific page.

    https://xtools.readthedocs.io/en/stable/api/user.html#top-edits

    :param project:
    :param username:
    :param namespace: one namespace (default is ``"0"``, the main one) or ``"all"`` for all of them.
    :param page_title: page title. For non-main pages you must either fill the ``namespace`` parameter or set it
      to a blank string and use the namespace-prefixed article title.
    :return:
    """
    path = base.build_path("/user/top_edits/{project}/{username}/{namespace}/{article}", (
        ("project", project, ""),
        ("username", username, ""),
        ("namespace", namespace, "0"),
        ("article", page_title, ""),
    ))
    return base.get(path)


def category_edit_counter(project: str, username: str, categories: Sequence[str],
                          start: Optional[date] = None,
                          end: Optional[date] = None) -> dict:
    """
    https://xtools.readthedocs.io/en/stable/api/user.html#category-edit-counter

    :param project:
    :param username:
    :param categories: Categories. The namespace prefix can be omitted.
    :param start:
    :param end:
    :return:
    """
    path = base.build_path("/user/category_editcount/{project}/{username}/{categories}/{start}/{end}", (
        ("project", project, ""),
        ("username", username, ""),
        ("categories", "|".join(categories), ""),
        ("start", start, ""),
        ("end", end, ""),
    ))
    return base.get(path)


def log_counts(project: str, username: str) -> dict:
    """
    Return counts of logged actions made by a user.

    https://xtools.readthedocs.io/en/stable/api/user.html#log-counts

    Example:

        >>> log_counts("en.wikipedia.org", "Jimbo_Wales")
        {
          "project": "en.wikipedia.org",
          "username": "Jimbo_Wales",
          "log_counts": {
            "block-block": 83,
            "block-unblock": 42,
            "create-create": 2,
            "delete-delete": 325,
            "delete-delete_redir": 1,
            "delete-event": 1,
            "delete-restore": 20,
            ...
          },
        }

    :param project:
    :param username:
    :return:
    """
    return base.get("/user/log_counts/%s/%s" % (project, username))


def namespace_totals(project: str, username: str) -> dict:
    """
    Return the edit count for each namespace with at least 1 edit.

    https://xtools.readthedocs.io/en/stable/api/user.html#namespace-totals

    :param project:
    :param username:
    :return:
    """
    return base.get("/user/namespace_totals/%s/%s" % (project, username))


def month_counts(project: str, username: str) -> dict:
    """
    Return the edit count of a user, grouped by namespace then year and month.

    https://xtools.readthedocs.io/en/stable/api/user.html#month-counts

    :param project:
    :param username:
    :return:
    """
    return base.get("/user/month_counts/%s/%s" % (project, username))


def time_card(project: str, username: str) -> dict:
    """
    Return the relative distribution of edits made by a user based on hour of the day and day of the week.

    Example:

        >>> time_card("en.wikipedia.org", "Jimbo_Wales")
        {
          "project": "en.wikipedia.org",
          "username": "Jimbo_Wales",
          "timecard": [
            {
              "day_of_week": 1,
              "hour": 0,
              "value": 83,
              "scale": 9
            },
            ...
          ],
        }

    https://xtools.readthedocs.io/en/stable/api/user.html#time-card

    :param project:
    :param username:
    :return:
    """
    return base.get("/user/timecard/%s/%s" % (project, username))


"""
Internal unit-testing utilities.
"""

import unittest

from xtools import base

TEST_URL_PREFIX = "m://x"


class TestCase(unittest.TestCase):
    def setUp(self):
        self.base_url = base.BASE_URL
        setattr(base, "BASE_URL", TEST_URL_PREFIX)

    def tearDown(self):
        setattr(base, "BASE_URL", self.base_url)


from setuptools import setup

setup(name="xtools",
      version="0.1.1",
      author="Baptiste Fontaine",
      author_email="b@ptistefontaine.fr",
      description="XTools API wrapper",
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      url="https://github.com/bfontaine/xtools",
      install_requires=['requests'],
      license="MIT",
      packages=['xtools'],
      # https://pypi.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Topic :: Software Development',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Operating System :: OS Independent',
      ])



from .page import article_info, prose, links, top_editors, assessments
from .project import (normalize_project, namespaces, page_assessments, page_assessments_configuration, automated_tools,
                      admins_and_user_groups, admin_statistics, patroller_statistics, steward_statistics)
from .quote import random_quote, single_quote, all_quotes
from .user import (simple_edit_count, number_of_pages_created, pages_created, pages_created_iter,
                   automated_edit_counter,  non_automated_edits, automated_edits,  edit_summaries, top_edits,
                   category_edit_counter, log_counts, namespace_totals, month_counts, time_card)

from .exceptions import BaseXToolsException, NotFound, TooManyEdits

__version__ = "0.1.1"
